@@ -233,12 +233,10 @@ static int hasher(struct kcapi_handle *handle, char *filename,
 			memset(compmd, 0, sizeof(compmd));
 			hex2bin(comphash, comphashlen, compmd, sizeof(compmd));
 			if ((comphashlen != (uint32_t)(ret * 2)) ||
-			    memcmp(compmd, md, ret)) {
+			    memcmp(compmd, md, ret))
 				ret = 1;
-			} else {
-
+			else
 				ret = 0;
-			}
 
 		} else {
 			bin2print(md, ret, filename, outfile);
@@ -299,7 +297,7 @@ static int hash_files(char *hashname, char *filename[], uint32_t files,
 		      const uint8_t *hmackey, uint32_t hmackeylen,
 		      int fipshmac)
 {
-	struct kcapi_handle handle;
+	struct kcapi_handle *handle;
 	uint32_t i = 0;
 	int ret = 0;
 	
@@ -310,7 +308,7 @@ static int hash_files(char *hashname, char *filename[], uint32_t files,
 		return -EFAULT;
 	}
 	if (hmackey) {
-		ret = kcapi_md_setkey(&handle, hmackey, hmackeylen);
+		ret = kcapi_md_setkey(handle, hmackey, hmackeylen);
 		if (ret) {
 			fprintf(stderr, "Setting HMAC key for %s failed (%d)\n",
 				hashname, ret);
@@ -337,14 +335,14 @@ static int hash_files(char *hashname, char *filename[], uint32_t files,
 			}
 			free(outfile);
 		}
-		ret = hasher(&handle, filename[i], NULL, 0, out);
+		ret = hasher(handle, filename[i], NULL, 0, out);
 		if (fipshmac)
 			fclose(out);
 		if (ret)
 			break;
 	}
 
-	kcapi_md_destroy(&handle);
+	kcapi_md_destroy(handle);
 	return ret;
 }
 
@@ -357,7 +355,7 @@ static int process_checkfile(char *hashname, char *checkfile, char *targetfile,
 {
 	FILE *file = NULL;
 	int ret = 0;
-	struct kcapi_handle handle;
+	struct kcapi_handle *handle;
 	/*
 	 * A file can have up to 4096 characters, so a complete line has at most
 	 * 4096 bytes (file name) + 128 bytes (SHA512 hex value) + 2 spaces +
@@ -372,7 +370,7 @@ static int process_checkfile(char *hashname, char *checkfile, char *targetfile,
 		return -EFAULT;
 	}
 	if (hmackey) {
-		ret = kcapi_md_setkey(&handle, hmackey, hmackeylen);
+		ret = kcapi_md_setkey(handle, hmackey, hmackeylen);
 		if (ret) {
 			fprintf(stderr, "Setting HMAC key for %s failed (%d)\n",
 				hashname, ret);
@@ -411,7 +409,7 @@ static int process_checkfile(char *hashname, char *checkfile, char *targetfile,
 				hashlen++;
 			} else {
 				char *filename = buf + i;
-				int r = hasher(&handle, filename, buf, hashlen,
+				int r = hasher(handle, filename, buf, hashlen,
 					       stdout);
 
 				if (r == 0) {
@@ -430,7 +428,7 @@ static int process_checkfile(char *hashname, char *checkfile, char *targetfile,
 
 		/* fipscheck does not have the filename in the check file */
 		if (!foundsep && targetfile) {
-			return hasher(&handle, targetfile,
+			return hasher(handle, targetfile,
 				      buf, hashlen - 1, stdout);
 		}
 
@@ -441,7 +439,7 @@ static int process_checkfile(char *hashname, char *checkfile, char *targetfile,
 out:
 	if (file)
 		fclose(file);
-	kcapi_md_destroy(&handle);
+	kcapi_md_destroy(handle);
 	return ret;
 
 }
@@ -473,9 +471,14 @@ static int fipscheck_self(char *hash,
 
 	fipsfile = fopen("/proc/sys/crypto/fips_enabled", "r");
 	if (!fipsfile) {
-		fprintf(stderr, "Cannot open fips_enabled file: %s\n",
-			strerror(errno));
-		return -EIO;
+		if (errno == ENOENT) {
+			/* FIPS support not enabled in kernel */
+			return 0;
+		} else {
+			fprintf(stderr, "Cannot open fips_enabled file: %s\n",
+				strerror(errno));
+			return -EIO;
+		}
 	}
 
 	n = fread((void *)fipsflag, 1, 1, fipsfile);
