@@ -1,7 +1,7 @@
 /*
  * Generic kernel crypto API user space interface library
  *
- * Copyright (C) 2014 - 2015, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2014 - 2016, Stephan Mueller <smueller@chronox.de>
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -40,6 +40,7 @@
 #include <fcntl.h>
 #include <sys/uio.h>
 #include <stdio.h>
+#include <stdarg.h>
 #include <unistd.h>
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -62,10 +63,10 @@
 		       * require consumer to be updated (as long as this number
 		       * is zero, the API is not considered stable and can
 		       * change without a bump of the major version) */
-#define MINVERSION 10 /* API compatible, ABI may change, functional
+#define MINVERSION 11 /* API compatible, ABI may change, functional
 		       * enhancements only, consumer can be left unchanged if
 		       * enhancements are not considered */
-#define PATCHLEVEL 2  /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 0  /* API / ABI compatible, no functional changes, no
 		       * enhancements, bug fixes only */
 
 /* remove once in if_alg.h */
@@ -99,6 +100,133 @@
 #define ALG_MAX_PAGES 16
 #endif
 
+/************************************************************
+ * Declarations for opague data structures
+ ************************************************************/
+
+/**
+ * Information obtained for different ciphers during handle init time
+ * using the NETLINK_CRYPTO interface.
+ * @blocksize block size of cipher (hash, symmetric, AEAD)
+ * @ivsize size of IV of cipher (symmetric, AEAD)
+ * @hash_digestsize size of message digest (hash)
+ * @blk_min_keysize minimum key size (symmetric)
+ * @blk_max_keysize maximum key size (symmetric)
+ * @aead_maxauthsize maximum authentication tag size (AEAD)
+ * @rng_seedsize seed size (RNG)
+ */
+struct kcapi_cipher_info {
+	/* generic */
+	uint32_t blocksize;
+	uint32_t ivsize;
+	/* hash */
+	uint32_t hash_digestsize;
+	/* blkcipher */
+	uint32_t blk_min_keysize;
+	uint32_t blk_max_keysize;
+	/* aead */
+	uint32_t aead_maxauthsize;
+	/* rng */
+	uint32_t rng_seedsize;
+};
+
+/**
+ * Common data required for symmetric and AEAD ciphers
+ * @iv: IV with length of kcapi_cipher_info->ivsize - input
+ */
+struct kcapi_cipher_data {
+	const uint8_t *iv;
+};
+
+/**
+ * AEAD data
+ * @datalen: Length of plaintext / ciphertext data - input
+ * @data: Pointer to plaintext / ciphertext data - input / output (the length is
+ *	  calculated with: kcapi_skcipher_data->inlen -
+ *			   kcapi_aead_data->taglen - kcapi_aead_data->assoclen)
+ * @assoclen: Length of associated data - input
+ * @assoc: Pointer to associated data - input
+ * @taglen: Length of authentication tag - input
+ * @tag: Authentication tag - input for decryption, output for encryption
+ * @retlen: internal data -- number plaintext / ciphertext bytes returned by
+ *	    the read system call
+ */
+struct kcapi_aead_data {
+	uint32_t datalen;
+	uint8_t *data;
+	uint32_t assoclen;
+	uint8_t *assoc;
+	uint32_t taglen;
+	uint8_t *tag;
+};
+
+/**
+ * Cipher handle
+ * @tfmfd: Socket descriptor for AF_ALG
+ * @opfd: FD to open kernel crypto API TFM
+ * @pipes: vmplice/splice pipe pair
+ * @processed_sg: number of scatter/gather entries sent to the kernel
+ * @ciper: Common data for all ciphers
+ * @aead: AEAD cipher specific data
+ * @info: properties of ciphers
+ * @aio: AIO information
+ */
+struct kcapi_handle {
+	int tfmfd;
+	int opfd;
+	int pipes[2];
+	uint32_t processed_sg;
+	struct kcapi_cipher_data cipher;
+	struct kcapi_aead_data aead;
+	struct kcapi_cipher_info info;
+	struct kcapi_aio aio;
+};
+
+/************************************************************
+ * Logging logic
+ ************************************************************/
+static int kcapi_verbosity_level = LOG_ERR;
+
+static void kcapi_dolog(int severity, const char *fmt, ...)
+{
+	va_list args;
+	char msg[128];
+	char sev[16];
+
+	if (severity > kcapi_verbosity_level)
+		return;
+
+	memset(sev, 0, sizeof(sev));
+	memset(msg, 0, sizeof(msg));
+
+	va_start(args, fmt);
+	vsnprintf(msg, sizeof(msg) - 1, fmt, args);
+	va_end(args);
+
+	switch (severity) {
+	case LOG_DEBUG:
+		snprintf(sev, sizeof(sev), "Debug");
+		break;
+	case LOG_VERBOSE:
+		snprintf(sev, sizeof(sev), "Verbose");
+		break;
+	case LOG_WARN:
+		snprintf(sev, sizeof(sev), "Warning");
+		break;
+	case LOG_ERR:
+		snprintf(sev, sizeof(sev), "Error");
+		break;
+	default:
+		snprintf(sev, sizeof(sev), "Unknown");
+	}
+	fprintf(stderr, "libkcapi - %s: %s\n", sev, msg);
+}
+
+void kcapi_set_verbosity(enum kcapi_verbosity level)
+{
+	kcapi_verbosity_level = level;
+}
+
 /************************************************************
  * Internal logic
  ************************************************************/
@@ -113,11 +241,12 @@ static int _kcapi_common_accept(struct kcapi_handle *handle)
 		int errsv = 0;
 
 		errsv = errno;
-		perror("AF_ALG: accept failed");
+		kcapi_dolog(LOG_ERR, "AF_ALG: accept failed");
 		close(handle->tfmfd);
 		handle->tfmfd = -1;
 		return -errsv;
 	}
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: accept syscall successful");
 
 	return 0;
 }
@@ -198,6 +327,8 @@ static int32_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 
 	ret = sendmsg(handle->opfd, &msg, flags);
 	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: sendmsg syscall returned %d (errno: %d)",
+		    ret, errsv);
 
 	kcapi_memset_secure(buffer, 0, bufferlen);
 	free(buffer);
@@ -210,6 +341,7 @@ static inline int32_t _kcapi_common_send_data(struct kcapi_handle *handle,
 {
 	struct msghdr msg;
 	int32_t ret = 0;
+	int32_t errsv;
 
 	ret = _kcapi_common_accept(handle);
 	if (ret)
@@ -224,7 +356,11 @@ static inline int32_t _kcapi_common_send_data(struct kcapi_handle *handle,
 	msg.msg_iovlen = iovlen;
 
 	ret = sendmsg(handle->opfd, &msg, flags);
-	return (ret >= 0) ? ret : -errno;
+	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: sendmsg syscall returned %d (errno: %d)",
+		    ret, errsv);
+
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static inline int32_t _kcapi_common_vmsplice_iov(struct kcapi_handle *handle,
@@ -235,6 +371,7 @@ static inline int32_t _kcapi_common_vmsplice_iov(struct kcapi_handle *handle,
 	int32_t ret = 0;
 	uint32_t inlen = 0;
 	unsigned long i;
+	int32_t errsv;
 
 	ret = _kcapi_common_accept(handle);
 	if (ret)
@@ -254,13 +391,20 @@ static inline int32_t _kcapi_common_vmsplice_iov(struct kcapi_handle *handle,
 	ret = vmsplice(handle->pipes[1], iov, iovlen, SPLICE_F_GIFT|flags);
 	if (0 > ret)
 		return ret;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: vmsplice syscall returned %d (errno: %d)",
+		    ret, errno);
+
 	if ((uint32_t)ret != inlen) {
-		fprintf(stderr, "vmsplice: not all data received by kernel (data recieved: %ld -- data sent: %lu)\n",
+		kcapi_dolog(LOG_ERR, "vmsplice: not all data received by kernel (data recieved: %ld -- data sent: %lu)",
 			(long)ret, (unsigned long)inlen);
 		return -EFAULT;
 	}
 	ret = splice(handle->pipes[0], NULL, handle->opfd, NULL, ret, flags);
-	return (ret >= 0) ? ret : -errno;
+	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: splice syscall returned %d (errno: %d)",
+		    ret, errsv);
+
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static inline int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
@@ -289,10 +433,14 @@ static inline int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
 		} else {
 			ret = vmsplice(handle->pipes[1], &iov, 1,
 				       SPLICE_F_GIFT|flags);
+			kcapi_dolog(LOG_DEBUG, "AF_ALG: vmsplice syscall returned %d (errno: %d)",
+				    ret, errno);
 			if (0 > ret)
 				return ret;
 			ret = splice(handle->pipes[0], NULL, handle->opfd,
 				     NULL, ret, flags);
+			kcapi_dolog(LOG_DEBUG, "AF_ALG: splice syscall returned %d (errno: %d)",
+				    ret, errno);
 		}
 		if (0 > ret)
 			return ret;
@@ -310,7 +458,7 @@ static inline int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 {
 	struct msghdr msg;
 	int32_t ret = 0;
-	int errsv = 0;
+	int32_t errsv = 0;
 
 	ret = _kcapi_common_accept(handle);
 	if (ret)
@@ -326,6 +474,8 @@ static inline int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 
 	ret = recvmsg(handle->opfd, &msg, 0);
 	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: recvmsg syscall returned %d (errno: %d)",
+		    ret, errsv);
 
 #if 0
 	/*
@@ -355,22 +505,32 @@ static inline int32_t _kcapi_common_read_data(struct kcapi_handle *handle,
 					      uint8_t *out, uint32_t outlen)
 {
 	int32_t ret = 0;
+	int32_t errsv;
 
 	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
 	ret = read(handle->opfd, out, outlen);
-	return (ret >= 0) ? ret : -errno;
+	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: read syscall returned %d (errno: %d)",
+		    ret, errsv);
+
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static inline int _kcapi_common_setkey(struct kcapi_handle *handle,
 				       const uint8_t *key, uint32_t keylen)
 {
 	int ret = 0;
+	int errsv;
 
 	ret = setsockopt(handle->tfmfd, SOL_ALG, ALG_SET_KEY, key, keylen);
-	return (ret >= 0) ? ret : -errno;
+	errsv = errno;
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: sendmsg syscall returned %d (errno: %d)",
+		    ret, errsv);
+
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static int __kcapi_common_getinfo(struct kcapi_handle *handle,
@@ -414,36 +574,35 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 	else
 		strncpy(req.cru.cru_name, ciphername, strlen(ciphername));
 
-
 	/* talk to netlink socket */
 	sd =  socket(AF_NETLINK, SOCK_RAW, NETLINK_CRYPTO);
 	if (sd < 0) {
-		perror("Netlink error: cannot open netlink socket");
+		kcapi_dolog(LOG_ERR, "Netlink error: cannot open netlink socket");
 		return -errno;
 	}
 	memset(&nl, 0, sizeof(nl));
 	nl.nl_family = AF_NETLINK;
 	if (bind(sd, (struct sockaddr*)&nl, sizeof(nl)) < 0) {
 		errsv = errno;
-		perror("Netlink error: cannot bind netlink socket");
+		kcapi_dolog(LOG_ERR, "Netlink error: cannot bind netlink socket");
 		goto out;
 	}
 	/* sanity check that netlink socket was successfully opened */
 	addr_len = sizeof(nl);
 	if (getsockname(sd, (struct sockaddr*)&nl, &addr_len) < 0) {
 		errsv = errno;
-		perror("Netlink error: cannot getsockname");
+		kcapi_dolog(LOG_ERR, "Netlink error: cannot getsockname");
 		goto out;
 	}
 	if (addr_len != sizeof(nl)) {
 		errsv = errno;
-		fprintf(stderr, "Netlink error: wrong address length %d\n",
+		kcapi_dolog(LOG_ERR, "Netlink error: wrong address length %d",
 			addr_len);
 		goto out;
 	}
 	if (nl.nl_family != AF_NETLINK) {
 		errsv = errno;
-		fprintf(stderr, "Netlink error: wrong address family %d\n",
+		kcapi_dolog(LOG_ERR, "Netlink error: wrong address family %d",
 			nl.nl_family);
 		goto out;
 	}
@@ -459,7 +618,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 	msg.msg_iovlen = 1;
 	if (sendmsg(sd, &msg, 0) < 0) {
 		errsv = errno;
-		perror("Netlink error: sendmsg failed");
+		kcapi_dolog(LOG_ERR, "Netlink error: sendmsg failed");
 		goto out;
 	}
 	memset(buf,0,sizeof(buf));
@@ -471,17 +630,17 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 			if (errno == EINTR || errno == EAGAIN)
 				continue;
 			errsv = errno;
-			perror("Netlink error: netlink receive error");
+			kcapi_dolog(LOG_ERR, "Netlink error: netlink receive error");
 			goto out;
 		}
 		if (ret == 0) {
 			errsv = errno;
-			fprintf(stderr, "Netlink error: no data\n");
+			kcapi_dolog(LOG_ERR, "Netlink error: no data");
 			goto out;
 		}
 		if ((uint32_t)ret > sizeof(buf)) {
 			errsv = errno;
-			perror("Netlink error: received too much data\n");
+			kcapi_dolog(LOG_ERR, "Netlink error: received too much data");
 			goto out;
 		}
 		break;
@@ -504,7 +663,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		res_len -= NLMSG_SPACE(sizeof(*cru_res));
 	}
 	if (res_len < 0) {
-		fprintf(stderr, "Netlink error: nlmsg len %d\n", res_len);
+		kcapi_dolog(LOG_ERR, "Netlink error: nlmsg len %d", res_len);
 		goto out;
 	}
 
@@ -517,8 +676,8 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		rta = RTA_NEXT(rta, res_len);
 	}
 	if (res_len) {
-		fprintf(stderr, "Netlink error: unprocessed data %d\n",
-			res_len);
+		kcapi_dolog(LOG_ERR, "Netlink error: unprocessed data %d",
+			    res_len);
 		goto out;
 	}
 
@@ -528,6 +687,8 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 			(struct crypto_report_hash *) RTA_DATA(rta);
 		handle->info.hash_digestsize = rsh->digestsize;
 		handle->info.blocksize = rsh->blocksize;
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: hash with digestsize %u,  blocksize %u",
+			    rsh->digestsize, rsh->blocksize);
 	}
 	if (tb[CRYPTOCFGA_REPORT_BLKCIPHER]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_BLKCIPHER];
@@ -537,6 +698,9 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		handle->info.ivsize = rblk->ivsize;
 		handle->info.blk_min_keysize = rblk->min_keysize;
 		handle->info.blk_max_keysize = rblk->max_keysize;
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: block cipher with blocksize %u, ivsize %u, minimum keysize %u, maximum keysize %u",
+			    rblk->blocksize, rblk->ivsize, rblk->min_keysize,
+			    rblk->max_keysize);
 	}
 	if (tb[CRYPTOCFGA_REPORT_AEAD]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_AEAD];
@@ -545,13 +709,32 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		handle->info.blocksize = raead->blocksize;
 		handle->info.ivsize = raead->ivsize;
 		handle->info.aead_maxauthsize = raead->maxauthsize;
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: AEAD block cipher with blocksize %u, ivsize %u, maximum authentication size %u",
+			    raead->blocksize, raead->ivsize,
+			    raead->maxauthsize);
 	}
 	if (tb[CRYPTOCFGA_REPORT_RNG]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_RNG];
 		struct crypto_report_rng *rrng =
 			(struct crypto_report_rng *) RTA_DATA(rta);
 		handle->info.rng_seedsize = rrng->seedsize;
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: RNG cipher with seedsize %u",
+			    rrng->seedsize);
 	}
+	if (tb[CRYPTOCFGA_UNSPEC])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: unspecified data received");
+	if (tb[CRYPTOCFGA_PRIORITY_VAL])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: u32 value received");
+	if (tb[CRYPTOCFGA_REPORT_LARVAL])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: larval value received");
+	if (tb[CRYPTOCFGA_REPORT_COMPRESS])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: compression algorithm type received");
+	if (tb[CRYPTOCFGA_REPORT_CIPHER])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: simple cipher algorithm type received");
+	if (tb[CRYPTOCFGA_REPORT_AKCIPHER])
+		kcapi_dolog(LOG_DEBUG, "Get cipher info: asymmetric cipher algorithm type received");
+	kcapi_dolog(LOG_VERBOSE, "Get cipher info: all information for %s received from kernel",
+		    ciphername);
 
 	ret = 0;
 
@@ -580,9 +763,12 @@ static void _kcapi_aio_init(struct kcapi_handle *handle)
 		goto err;
 
 	if (io_setup(KCAPI_AIO_CONCURRENT, &handle->aio.aio_ctx) < 0) {
-		perror("io_setup error");
+		kcapi_dolog(LOG_ERR, "io_setup error");
 		goto err;
 	}
+
+	kcapi_dolog(LOG_VERBOSE, "asynchronoous I/O initialized");
+
 	return;
 
 err:
@@ -616,14 +802,22 @@ static inline void _kcapi_handle_destroy(struct kcapi_handle *handle)
 	memset(handle, 0, sizeof(struct kcapi_handle));
 }
 
-static int _kcapi_handle_init(struct kcapi_handle *handle, const char *type,
+static int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 			      const char *ciphername, uint32_t flags)
 {
 	struct sockaddr_alg sa;
+	struct kcapi_handle *handle;
 	int ret;
 	int errsv = 0;
+	char versionbuffer[50];
 
-	memset(handle, 0, sizeof(struct kcapi_handle));
+	kcapi_versionstring(versionbuffer, sizeof(versionbuffer));
+	kcapi_dolog(LOG_VERBOSE, "%s - initializing cipher operation with kernel",
+		    versionbuffer);
+
+	handle = calloc(1, sizeof(struct kcapi_handle));
+	if (!handle)
+		return -ENOMEM;
 
 	memset(&sa, 0, sizeof(sa));
 	sa.salg_family = AF_ALG;
@@ -631,31 +825,39 @@ static int _kcapi_handle_init(struct kcapi_handle *handle, const char *type,
 	snprintf((char *)sa.salg_name, sizeof(sa.salg_name),"%s", ciphername);
 
 	handle->tfmfd = socket(AF_ALG, SOCK_SEQPACKET, 0);
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: socket syscall %s (errno: %d)",
+		    (handle->tfmfd == -1) ? "failed": "passed", errno);
 	if (handle->tfmfd == -1)
 		return -EOPNOTSUPP;
 
 	if (bind(handle->tfmfd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
 		errsv = errno;
-		perror("AF_ALG: bind failed");
+		kcapi_dolog(LOG_ERR, "AF_ALG: bind failed (errno: %d)",
+			    errsv);
 		close(handle->tfmfd);
 		handle->tfmfd = -1;
 		return -errsv;
 	}
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: bind syscall passed (errno: %d)",
+		    errno);
 
 	handle->opfd = -1;
 
 	ret = pipe(handle->pipes);
 	if (ret) {
 		errsv = errno;
+		kcapi_dolog(LOG_DEBUG, "AF_ALG: pipe syscall failed (errno: %d)",
+			    errsv);
 		close(handle->tfmfd);
 		return -errsv;
 	}
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: pipe syscall passed");
 
 	ret = _kcapi_common_getinfo(handle, ciphername);
 	if (ret) {
 		errsv = errno;
-		fprintf(stderr, "NETLINK_CRYPTO: cannot obtain cipher information for %s (is required crypto_user.c patch missing? see documentation)\n",
-		       ciphername);
+		kcapi_dolog(LOG_ERR, "NETLINK_CRYPTO: cannot obtain cipher information for %s (is required crypto_user.c patch missing? see documentation)",
+			    ciphername);
 		_kcapi_handle_destroy(handle);
 	}
 
@@ -664,6 +866,10 @@ static int _kcapi_handle_init(struct kcapi_handle *handle, const char *type,
 	else
 		handle->aio.skcipher_aio_disable = 1;
 
+	kcapi_dolog(LOG_VERBOSE, "communication for %s with kernel initialized", ciphername);
+
+	*caller = handle;
+
 	return (errsv) ? -errsv : ret;
 }
 
@@ -714,7 +920,7 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 	return 0;
 }
 
-int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername,
+int kcapi_cipher_init(struct kcapi_handle **handle, const char *ciphername,
 		      uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "skcipher", ciphername, flags);
@@ -722,7 +928,10 @@ int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername,
 
 void kcapi_cipher_destroy(struct kcapi_handle *handle)
 {
+	if (!handle)
+		return;
 	_kcapi_handle_destroy(handle);
+	kcapi_memset_secure(handle, 0, sizeof(struct kcapi_handle));
 }
 
 int kcapi_cipher_setkey(struct kcapi_handle *handle,
@@ -737,7 +946,7 @@ static inline int32_t _kcapi_aio_read(struct kcapi_handle *handle,
 	(void)handle;
 	(void)out;
 	(void)outlen;
-	fprintf(stderr, "AIO support not complete\n");
+	kcapi_dolog(LOG_ERR, "AIO support not complete");
 	return -EOPNOTSUPP;
 #if 0
 	struct iocb *cb = NULL;
@@ -783,8 +992,8 @@ static int32_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
 	int32_t ret = 0;
 
 	if (!in || !inlen || !out || !outlen) {
-		fprintf(stderr,
-			"Symmetric Encryption: Empty plaintext or ciphertext buffer provided\n");
+		kcapi_dolog(LOG_ERR,
+			    "Symmetric Encryption: Empty plaintext or ciphertext buffer provided");
 		return -EINVAL;
 	}
 
@@ -865,9 +1074,9 @@ int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 
 	/* require properly sized output data size */
 	if (outlen < ((inlen + bs - 1) / bs * bs)) {
-		fprintf(stderr,
-			"Symmetric Encryption: Ciphertext buffer (%lu) is not plaintext buffer (%lu) rounded up to multiple of block size %u\n",
-			(unsigned long) outlen, (unsigned long)inlen, bs);
+		kcapi_dolog(LOG_ERR,
+			    "Symmetric Encryption: Ciphertext buffer (%lu) is not plaintext buffer (%lu) rounded up to multiple of block size %u",
+			    (unsigned long) outlen, (unsigned long)inlen, bs);
 		return -EINVAL;
 	}
 
@@ -883,16 +1092,16 @@ int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 {
 	/* require properly sized output data size */
 	if (inlen % handle->info.blocksize) {
-		fprintf(stderr,
-			"Symmetric Decryption: Ciphertext buffer is not multiple of block size %u\n",
-			handle->info.blocksize);
+		kcapi_dolog(LOG_ERR,
+			    "Symmetric Decryption: Ciphertext buffer is not multiple of block size %u",
+			    handle->info.blocksize);
 		return -EINVAL;
 	}
 
 	if (outlen < inlen) {
-		fprintf(stderr,
-			"Symmetric Decryption: Plaintext buffer (%lu) is smaller as ciphertext buffer (%lu)\n",
-			(unsigned long)outlen, (unsigned long)inlen);
+		kcapi_dolog(LOG_ERR,
+			    "Symmetric Decryption: Plaintext buffer (%lu) is smaller as ciphertext buffer (%lu)",
+			    (unsigned long)outlen, (unsigned long)inlen);
 		return -EINVAL;
 	}
 
@@ -933,8 +1142,8 @@ int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
 			       struct iovec *iov, uint32_t iovlen)
 {
 	if (!iov || !iovlen) {
-		fprintf(stderr,
-			"Symmetric operation: No buffer for output data provided\n");
+		kcapi_dolog(LOG_ERR,
+			    "Symmetric operation: No buffer for output data provided");
 		return -EINVAL;
 	}
 	return _kcapi_common_recv_data(handle, iov, iovlen);
@@ -950,7 +1159,7 @@ uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle)
 	return handle->info.blocksize;
 }
 
-int kcapi_aead_init(struct kcapi_handle *handle, const char *ciphername,
+int kcapi_aead_init(struct kcapi_handle **handle, const char *ciphername,
 		    uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "aead", ciphername, flags);
@@ -990,10 +1199,10 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
 			uint8_t **tag, uint32_t *taglen)
 {
 	if (encdatalen <  handle->aead.taglen + handle->aead.assoclen) {
-		fprintf(stderr, "Result of encryption operation (%lu) is smaller than tag and AAD length (%lu)\n",
-			(unsigned long)encdatalen,
-			(unsigned long)handle->aead.taglen +
-			(unsigned long)handle->aead.assoclen);
+		kcapi_dolog(LOG_ERR, "Result of encryption operation (%lu) is smaller than tag and AAD length (%lu)",
+			    (unsigned long)encdatalen,
+			    (unsigned long)handle->aead.taglen +
+			    (unsigned long)handle->aead.assoclen);
 		return;
 	}
 	if (aad) {
@@ -1021,16 +1230,16 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 
 	/* require properly sized output data size */
 	if (outlen < inlen) {
-		fprintf(stderr,
-			"AEAD Encryption: Ciphertext buffer (%u) is smaller than plaintext buffer (%u)\n",
-			outlen, inlen);
+		kcapi_dolog(LOG_ERR,
+			    "AEAD Encryption: Ciphertext buffer (%u) is smaller than plaintext buffer (%u)",
+			    outlen, inlen);
 		return -EINVAL;
 	}
 
 	if (inlen > (uint32_t)(sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES)) {
-		fprintf(stderr,
-			"AEAD Encryption: Plaintext buffer (%u) is larger than maximum chunk size (%lu)\n",
-			inlen, sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES);
+		kcapi_dolog(LOG_ERR,
+			    "AEAD Encryption: Plaintext buffer (%u) is larger than maximum chunk size (%lu)",
+			    inlen, sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES);
 		return -EMSGSIZE;
 	}
 
@@ -1073,16 +1282,16 @@ int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 
 	/* require properly sized output data size */
 	if (outlen < inlen) {
-		fprintf(stderr,
-			"AEAD Decryption: Plaintext buffer (%u) is smaller than ciphertext buffer (%u)\n",
-			outlen, inlen);
+		kcapi_dolog(LOG_ERR,
+			    "AEAD Decryption: Plaintext buffer (%u) is smaller than ciphertext buffer (%u)",
+			    outlen, inlen);
 		return -EINVAL;
 	}
 
 	if (inlen > (uint32_t)(sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES)) {
-		fprintf(stderr,
-			"AEAD Decryption: Ciphertext buffer (%u) is larger than maximum chunk size (%lu)\n",
-			inlen, sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES);
+		kcapi_dolog(LOG_ERR,
+			    "AEAD Decryption: Ciphertext buffer (%u) is larger than maximum chunk size (%lu)",
+			    inlen, sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES);
 		return -EMSGSIZE;
 	}
 
@@ -1157,8 +1366,8 @@ int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			     struct iovec *iov, uint32_t iovlen)
 {
 	if (!iov) {
-		fprintf(stderr,
-			"AEAD operation: No buffer for output data provided\n");
+		kcapi_dolog(LOG_ERR,
+			    "AEAD operation: No buffer for output data provided");
 		return -EINVAL;
 	}
 
@@ -1210,7 +1419,7 @@ int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
 	return 0;
 }
 
-int kcapi_md_init(struct kcapi_handle *handle, const char *ciphername,
+int kcapi_md_init(struct kcapi_handle **handle, const char *ciphername,
 		  uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "hash", ciphername, flags);
@@ -1233,12 +1442,10 @@ static inline int32_t _kcapi_md_update(struct kcapi_handle *handle,
 	int32_t ret = 0;
 
 	/* zero buffer length cannot be handled via splice */
-	/* TODO check that heuristic for sendmsg is appropriate */
-	if(len == 0 /* < (1<<15) */) {
+	if (len < (1<<15)) {
 		ret = _kcapi_common_accept(handle);
 		if (ret)
 			return ret;
-
 		ret = send(handle->opfd, buffer, len, MSG_MORE);
 	} else {
 		ret = _kcapi_common_vmsplice_chunk(handle, buffer, len,
@@ -1262,10 +1469,9 @@ static int32_t _kcapi_md_final(struct kcapi_handle *handle,
 	struct iovec iov;
 
 	if (!buffer || !len) {
-		fprintf(stderr,
-			"Message digest: output buffer too small (seen %lu - required %u)\n",
-			(unsigned long)len,
-			handle->info.hash_digestsize);
+		kcapi_dolog(LOG_ERR,
+			    "Message digest: output buffer too small (seen %lu - required %u)",
+			    (unsigned long)len,	handle->info.hash_digestsize);
 		return -EINVAL;
 	}
 
@@ -1302,7 +1508,7 @@ uint32_t kcapi_md_blocksize(struct kcapi_handle *handle)
 	return handle->info.blocksize;
 }
 
-int kcapi_rng_init(struct kcapi_handle *handle, const char *ciphername,
+int kcapi_rng_init(struct kcapi_handle **handle, const char *ciphername,
 		   uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "rng", ciphername, flags);
@@ -1342,7 +1548,7 @@ int32_t kcapi_rng_generate(struct kcapi_handle *handle,
 	return out;
 }
 
-int kcapi_akcipher_init(struct kcapi_handle *handle, const char *ciphername,
+int kcapi_akcipher_init(struct kcapi_handle **handle, const char *ciphername,
 			uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "akcipher", ciphername, flags);
@@ -1453,8 +1659,8 @@ int32_t kcapi_akcipher_stream_op(struct kcapi_handle *handle,
 			         struct iovec *iov, uint32_t iovlen)
 {
 	if (!iov || !iovlen) {
-		fprintf(stderr,
-			"Symmetric operation: No buffer for output data provided\n");
+		kcapi_dolog(LOG_ERR,
+			    "Asymmetric operation: No buffer for output data provided");
 		return -EINVAL;
 	}
 	return _kcapi_common_recv_data(handle, iov, iovlen);
